import json
from typing import Optional, Dict, Any

from sageai.config import get_config, set_config
from sageai.services.default_vectordb_service import DefaultVectorDBService
from sageai.services.openai_service import OpenAIService
from sageai.types.abstract_vectordb import AbstractVectorDB
from sageai.types.function import Function
from sageai.utils import generate_functions_map
from sageai.utils.inspection_utilities import get_input_parameter_type


class SageAI:
    def __init__(
        self,
        *,
        openai_key: str,
        functions_directory: Optional[str] = None,
        function_calling_model: Optional[str] = None,
        embeddings_model: Optional[str] = None,
        vectordb: Optional[AbstractVectorDB] = None,
    ):
        config_args = {"openai_key": openai_key}

        if functions_directory is not None:
            config_args["functions_directory"] = functions_directory
        if function_calling_model is not None:
            config_args["function_calling_model"] = function_calling_model
        if embeddings_model is not None:
            config_args["embeddings_model"] = embeddings_model
        set_config(**config_args)
        self.config = get_config()
        self.openai = OpenAIService()
        self.function_map = generate_functions_map(self.config.functions_directory)
        self.vectordb = (
            vectordb(function_map=self.function_map)
            if vectordb
            else DefaultVectorDBService(function_map=self.function_map)
        )

    def chat(self, *, message: str, options: Optional[Dict] = None) -> str:
        """
        High-level function that calls the vector database, OpenAI, and the function, and returns
        the result.
        @param message: the message
        @param options:
        @return:
        """
        if options is None:
            options = {}

        top_functions = self.get_top_n_functions(
            message=message, k=options.get("k") or 5
        )
        openai_result = self.openai.chat(
            functions=top_functions,
            message=message,
        )

        if "function_call" not in openai_result:
            raise Exception("No function call found in OpenAI response.")

        function_to_call = openai_result["function_call"]["name"]
        function_args = json.loads(openai_result["function_call"]["arguments"])
        function_response = self.run_function(
            name=function_to_call,
            args=function_args,
        )
        return function_response

    def get_top_n_functions(self, *, message: str, k: int):
        return self.vectordb.search(query=message, k=k)

    def run_function(self, *, name: str, args: Dict[str, Any]):
        func = self.function_map[name]
        function_input_type = get_input_parameter_type(func.function)
        func_args = function_input_type(**args)
        return func(func_args)

    def index(self):
        self.vectordb.index()



import os
from importlib import util
from logging import Logger
from types import ModuleType
from typing import List

from sageai.types.function import Function
from sageai.utils.logger import get_logger


def load_module_from_file(filename: str, filepath: str) -> ModuleType:
    spec = util.spec_from_file_location(filename, filepath)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_functions_directories(
    logger: Logger,
    functions_directory_path: str = None,
) -> List[str]:
    logger.info(
        f"Getting functions directories from {functions_directory_path}",
    )

    function_directories = [
        dirpath
        for dirpath, dirnames, filenames in os.walk(functions_directory_path)
        if "function.py" in filenames
    ]
    logger.info(f"Found {len(function_directories)} function directories")
    return sorted(function_directories)


def generate_functions_map(functions_directory_path: str) -> dict[str, Function]:
    logger = get_logger()
    available_functions = {}

    logger.info("Generating function map")
    functions_directory = get_functions_directories(logger, functions_directory_path)

    for dirpath in functions_directory:
        folder_name = os.path.basename(dirpath)
        function_file = os.path.join(dirpath, "function.py")
        function_module = load_module_from_file(folder_name, function_file)

        if not hasattr(function_module, "function"):
            raise Exception(
                f"Function {folder_name} does not have a function attribute",
            )

        available_functions[function_module.function.name] = function_module.function

    if len(available_functions) == 0:
        raise Exception("No functions found")
    logger.info(f"Function map generated with {len(available_functions)} functions")
    available_functions = dict(sorted(available_functions.items()))

    return available_functions



from typing import Optional

from pydantic import BaseModel, ValidationError, Field

from sageai.utils import format_config_args


class Config(BaseModel):
    openai_key: str
    functions_directory: Optional[str] = Field(
        "functions", description="The directory of functions."
    )
    function_calling_model: Optional[str] = Field(
        "gpt-3.5-turbo-0613",
        description="The OpenAI model to use for function calling.",
    )
    embeddings_model: Optional[str] = Field(
        "text-embedding-ada-002", description="The OpenAI model to use for embeddings."
    )


_config = Config(openai_key="OPENAI_API_KEY")  # TODO change this


def set_config(**kwargs):
    """Set the configuration parameters."""
    global _config
    try:
        kwargs = format_config_args(kwargs)
        _config = Config(**kwargs)
        return _config
    except ValidationError as e:
        raise ValidationError(f"Invalid configuration: {e}")


def get_config():
    """Retrieve the configuration."""
    return _config



import inspect
import os


def get_user_package_path():
    caller_frame = inspect.stack()[4]
    caller_module = inspect.getmodule(caller_frame[0])

    if caller_module is None or not hasattr(caller_module, "__file__"):
        raise ValueError("Cannot determine the path of the caller's module")

    module_path = os.path.abspath(caller_module.__file__)
    return os.path.dirname(module_path)


def format_config_args(args):
    if not os.path.isabs(args["functions_directory"]) and not args[
        "functions_directory"
    ].startswith("./"):
        base_path = get_user_package_path()
        args["functions_directory"] = os.path.join(
            base_path, args["functions_directory"]
        )
    elif args["functions_directory"].startswith("./"):
        base_path = get_user_package_path()
        # Strip './' from the beginning and then join with the base path
        relative_path = args["functions_directory"][2:]
        args["functions_directory"] = os.path.join(base_path, relative_path)
    return args



from typing import Any, Dict, List

from qdrant_client import QdrantClient
from qdrant_client.http.models import models

from sageai.services.openai_service import OpenAIService
from sageai.types.abstract_vectordb import AbstractVectorDB
from sageai.types.function import Function


class DefaultVectorDBService(AbstractVectorDB):
    def __init__(self, function_map: dict[str, Function]):
        from sageai.config import get_config

        self.client = QdrantClient(":memory:")
        self.openai = OpenAIService()
        self.config = get_config()
        self.collection = "functions"
        self.function_map = function_map

    def index(self):
        self.client.recreate_collection(
            collection_name=self.collection,
            vectors_config=models.VectorParams(
                size=self.openai.get_embeddings_size(),
                distance=models.Distance.COSINE,
            ),
        )

        def format_func_embedding(func: Function) -> str:
            embedding_text = func.name.replace("_", " ") + " - " + func.description
            return self.format_query(query=embedding_text)

        records = [
            models.Record(
                id=idx,
                vector=self.openai.create_embeddings(text=format_func_embedding(func)),
                payload={"func_name": func_name},
            )
            for idx, (func_name, func) in enumerate(self.function_map.items())
        ]
        self.client.upload_records(
            collection_name=self.collection,
            records=records,
        )

    def embed(self, query: str):
        return self.openai.create_embeddings(query)

    def format_query(self, *, query: str):
        return query.lower()

    def search(self, query: str, k: int) -> List[Dict[str, Any]]:
        formatted_query = self.format_query(query=query)
        query_embedding = self.openai.create_embeddings(text=formatted_query)
        hits = self.client.search(
            collection_name=self.collection,
            query_vector=query_embedding,
            limit=k,
        )
        func_names = [hit.payload["func_name"] for hit in hits]
        potential_functions = [
            self.function_map[func_name].parameters for func_name in func_names
        ]
        return potential_functions



from logging import getLogger
from typing import Any, List, Dict

import openai

from sageai.config import get_config


class OpenAIService:
    def __init__(self):
        self.logger = getLogger(__name__)

        config = get_config()
        self.api_key = config.openai_key
        self.function_calling_model = config.function_calling_model
        self.embeddings_model = config.embeddings_model

        openai.api_key = self.api_key
        self.openai = openai

    def create_embeddings(self, text: str) -> List[float]:
        response = self.openai.Embedding.create(
            input=text,
            model=self.embeddings_model,
        )
        embeddings = response["data"][0]["embedding"]
        return embeddings

    def chat(
        self,
        functions: List[dict[str, Any]],
        message: str,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        response = openai.ChatCompletion.create(
            model=kwargs["model"] if "model" in kwargs else self.function_calling_model,
            messages=[dict(role="user", content=message)],
            functions=functions,
            **kwargs,
        )
        response_message = response["choices"][0]["message"]
        return response_message

    @staticmethod
    def get_embeddings_size():
        return 1536



from abc import ABC, abstractmethod
from typing import List

from sageai.types.function import Function


class AbstractVectorDB(ABC):
    def __init__(self, function_map: dict[str, Function]):
        pass

    @abstractmethod
    def index(self) -> None:
        """Indexes the vector db based on the functions directory."""
        pass

    def embed(self, *, query: str):
        """Embeds a query."""
        pass

    def format_query(self, *, query: str) -> str:
        """Formats a query for embedding."""
        pass

    @abstractmethod
    def search(self, *, query: str, n: int) -> List[str]:
        """Search vector db based on a query and return top n function names."""
        pass



from typing import Any, Callable, Dict, List, Type

from pydantic import BaseModel

from sageai.utils.inspection_utilities import get_input_parameter_type
from sageai.utils.model_utilities import is_enum_array, get_enum_array_values, is_enum, get_possible_values, is_array, \
    get_array_item_type, is_optional


class Function(BaseModel):
    # required
    function: Callable
    description: str

    # generated from input
    name: str
    parameters: Dict[str, Any]
    input_type: Type[BaseModel]

    def __init__(
        self,
        function: Callable,
        description: str,
    ) -> None:
        name = function.__name__
        input_parameter_type = get_input_parameter_type(function)
        formatted_parameters = self._format_parameters(
            input_parameter_type=input_parameter_type,
            name=name,
            description=description,
        )

        super().__init__(
            function=function,
            description=description,
            name=name,
            parameters=formatted_parameters,
            input_type=input_parameter_type,
        )

    @staticmethod
    def _format_parameters(input_parameter_type, name: str, description: str):
        parameters = input_parameter_type.schema()
        new_schema = {
            "name": name,
            "description": description,
            "parameters": {
                "type": "object",
                "properties": {},
                "required": [],
            },
        }

        for key, value in parameters["properties"].items():
            if is_enum_array(input_parameter_type, key):
                enum_values = get_enum_array_values(input_parameter_type, key)
                enum_type = "integer" if isinstance(enum_values[0], int) else "string"
                property_schema = {
                    "type": "array",
                    "description": value.get("description", ""),
                    "items": {"type": enum_type, "enum": enum_values},
                }
            elif is_enum(input_parameter_type, key):
                property_schema = {
                    "type": "string",
                    "description": value["description"],
                    "enum": get_possible_values(input_parameter_type, key),
                }
            elif is_array(input_parameter_type, key):
                property_schema = {
                    "type": "array",
                    "description": value.get("description", ""),
                    "items": {
                        "type": get_array_item_type(input_parameter_type, key),
                    },
                }
            else:
                property_type = value.get("type")
                if property_type is None:
                    raise Exception("Property type is None")
                property_schema = {
                    "type": property_type,
                    "description": value.get("description", ""),
                }

            new_schema["parameters"]["properties"][key] = property_schema

            # Check if the property is required
            if not is_optional(input_parameter_type, key):
                new_schema["parameters"]["required"].append(key)

        return new_schema

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def __str__(self):
        return f"{self.name}: {self.description}"



from enum import Enum
from types import ModuleType
from typing import Any, List, Optional, Type, Union, get_args, get_origin

from pydantic import BaseModel


def find_pydantic_model(
    module: ModuleType,
    keyword: str,
) -> Type[BaseModel] | None:
    models = [
        cls
        for name, cls in vars(module).items()
        if isinstance(cls, type)
           and issubclass(cls, BaseModel)
           and name.endswith(
            keyword,
        )
    ]
    return models[0] if models else None


def get_input_model(module: ModuleType) -> Type[BaseModel]:
    return find_pydantic_model(module, keyword="Input")


def get_output_data_model(module: ModuleType) -> Type[BaseModel]:
    return find_pydantic_model(module, keyword="Data")


def is_enum(model: Type[BaseModel], key: str) -> bool:
    if key in model.__annotations__:
        annotation = model.__annotations__[key]
        # Check if annotation is a class
        if isinstance(annotation, type):
            return issubclass(annotation, Enum)
        # Handle Optional[Enum] and List[Enum]
        origin = get_origin(annotation)
        if origin in {Optional, List, Union}:
            args = get_args(annotation)
            if args and isinstance(args[0], type):
                return issubclass(args[0], Enum)
    return False


def get_possible_values(model: Type[BaseModel], key: str) -> List[str] | List[int]:
    if is_enum(model, key) or is_enum_array(model, key):
        enum_type = model.__annotations__[key]
        if get_origin(enum_type) in {Optional, Union, List}:
            enum_type = get_args(enum_type)[0]
        return [e.value for e in enum_type]
    return []


def is_optional(model: Type[BaseModel], key: str) -> bool:
    annotation = model.__annotations__[key]
    return type(None) in get_args(annotation)


def is_array(model: Type[BaseModel], key: str) -> bool:
    if key in model.__annotations__:
        annotation = model.__annotations__[key]
        if get_origin(annotation) is list:
            return True

        if get_origin(annotation) in {Optional, Union}:
            args = get_args(annotation)
            return any(get_origin(arg) is list for arg in args)
    return False


def get_array_item_type(model: Type[BaseModel], key: str) -> str | None:
    type_map = {"str": "string", "int": "integer"}
    if not is_array(model, key):
        raise Exception(f"{key} is not an array")
    field = model.__annotations__[key]
    if get_origin(field) is list:
        return type_map[get_args(field)[0].__name__]
    elif get_origin(field) in {Optional, Union}:
        args = get_args(field)
        for arg in args:
            if get_origin(arg) is list:
                return type_map[get_args(arg)[0].__name__]
    return None


def is_enum_array(model: Type[BaseModel], key: str) -> bool:
    enum_type = get_array_item_enum_type(model, key)
    return issubclass(enum_type, Enum) if enum_type is not None else False


def get_enum_type_from_annotation(field) -> Type[Enum] | None:
    enum_type = None
    if get_origin(field) is list:
        enum_type = get_args(field)[0]
    elif get_origin(field) in {Optional, Union}:
        args = get_args(field)
        for arg in args:
            if get_origin(arg) is list:
                enum_type = get_args(arg)[0]
    if enum_type and issubclass(enum_type, Enum):
        return enum_type
    return None


def get_array_item_enum_type(model: Type[BaseModel], key: str) -> Type | None:
    if not is_array(model, key):
        return None
    field = model.__annotations__[key]
    return get_enum_type_from_annotation(field)


def get_enum_array_values(
    model: Type[BaseModel],
    key: str,
) -> List[str] | List[int]:
    if not is_enum_array(model, key):
        raise Exception(f"{key} is not an array of Enums")
    field = model.__annotations__[key]
    enum_type = get_enum_type_from_annotation(field)
    if enum_type:
        return [e.value for e in enum_type.__members__.values()]
    return []


def get_default_value(model: Type[BaseModel], key: str) -> Any:
    if is_enum(model, key):
        default_value = model.__fields__[key].default
        if is_enum(model, key) and isinstance(default_value, Enum):
            return default_value.value
        return default_value

    if key in model.__fields__:
        default_value = model.__fields__[key].default
        if default_value == "None" or default_value is None:
            return None
        return default_value

    return None



import logging


def get_logger():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter("%(asctime)s | [%(levelname)s] %(message)s")

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)

    logger.addHandler(console_handler)

    return logger



import inspect
from typing import Callable, Type

from pydantic import BaseModel


def get_input_parameter_type(function: Callable) -> Type[BaseModel]:
    signature = inspect.signature(function)
    for name, param in signature.parameters.items():
        if "Input" in str(param.annotation):
            return param.annotation
    raise Exception(f"Could not find Input parameter type from function arguments.")
